/**
 * ForwardServerClientThread handles the clients of Nakov Forward Server. It
 * finds suitable server from the server pool, connects to it and starts
 * the TCP forwarding between given client and its assigned server. After
 * the forwarding is failed and the two threads are stopped, closes the sockets.
 */

import java.net.Socket;
import java.net.SocketException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;


public class ForwardServerClientThread extends Thread
{

	private NakovForwardServer mNakovForwardServer = null;
	private Socket mClientSocket = null;
	private Socket mServerSocket = null;
	private boolean mBothConnectionsAreAlive = false;
	private String mServerHost = null;
	private String mServerPort = null;


	/**
	 * Creates a client thread for handling clients of NakovForwardServer.
	 * A client socket should be connected and passed to this constructor.
	 * A server socket is created later by run() method.
	 */
	public ForwardServerClientThread(NakovForwardServer aNakovForwardServer, Socket aClientSocket)
	{
		mNakovForwardServer = aNakovForwardServer;
		mClientSocket = aClientSocket;
	}


	/**
	 * Obtains a destination server socket to some of the servers in the list.
	 * Starts two threads for forwarding : "client in <--> dest server out" and
	 * "dest server in <--> client out", waits until one of these threads stop
	 * due to read/write failure or connection closure. Closes opened connections.
	 */
	public void run()
	{
		try {
			// Create a new socket connection to one of the servers from the list
			mServerSocket = createServerSocket();
			if (mServerSocket==null) {  // If all the servers are down
				try { mClientSocket.close(); } catch (IOException e) {}
				return;
			}

			// Obtain input and output streams of server and client
			InputStream clientIn = mClientSocket.getInputStream();
			OutputStream clientOut = mClientSocket.getOutputStream();
			InputStream serverIn = mServerSocket.getInputStream();
			OutputStream serverOut = mServerSocket.getOutputStream();

			String clientHostPort = mClientSocket.getInetAddress().getHostAddress() + ":" + mClientSocket.getPort();
			String serverHostPort = mServerHost + ":" + mServerPort;
			mNakovForwardServer.log("TCP Forwarding  " + clientHostPort + " <--> " + serverHostPort + "  started.");

			// Start forwarding of socket data between server and client
			ForwardThread clientForward = new ForwardThread(this, clientIn, serverOut);
			ForwardThread serverForward = new ForwardThread(this, serverIn, clientOut);
			mBothConnectionsAreAlive = true;
			clientForward.start();
			serverForward.start();

			// Wait while the server and client threads are alive
			while (mBothConnectionsAreAlive) {
				try {
					sleep(1000); // Check every second if the server and client threads are alive
				} catch (InterruptedException ie) { }
			}

			mNakovForwardServer.log("TCP Forwarding  " + clientHostPort + " <--> " + serverHostPort + "  stopped.");

		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}


	/**
	 * connectionBroken() method is called by forwarding child threads to notify
	 * this thread (their parent thread) that one of the connections (server or client)
	 * is broken (a read/write failure occured). This method disconnects both server
	 * and client sockets causing both threads to stop forwarding.
	 */
	public synchronized void connectionBroken()
	{
		if (mBothConnectionsAreAlive) {
			// One of the connections is broken. Close the other connection and stop forwarding
			// Closing these socket connections will close their input/output streams of course
			try { mServerSocket.close(); } catch (IOException e) {}
			try { mClientSocket.close(); } catch (IOException e) {}
			mBothConnectionsAreAlive = false;
		}
	}


	/**
	 * @return a new socket connected to some of the servers in the destination
	 * servers list. Sequentially a connection to each of the alive threads is
	 * tried to be established. If connecting to first alive server fail, is it
	 * marked as dead and next alive server is tried. If all the servers are dead,
	 * null is returned. Thus if at least one server is alive, a connection will
	 * be established (of course after some delay) and the system will not fail
	 * (it is fault telerant). Dead servers can be marked as alive if revived, but
	 * this is done later by check alive thread.
	 */
	private Socket createServerSocket()
	throws IOException
	{
		// To be implemented better - to do "load balancing"
		NakovForwardServer.ServerDescription[] servers = mNakovForwardServer.getServersList();
		for (int i=0; i<servers.length; i++) {
			if (servers[i].isAlive) {
				try {
					Socket socket = new Socket(servers[i].host, servers[i].port);
					mServerHost = servers[i].host;
					mServerPort = "" + servers[i].port;
					return socket;
				} catch (IOException ioe) {
					servers[i].isAlive = false;
				}
			}
		}
		return null;
	}

}
